from dataclasses import dataclass

# TODO: Move all the standardized git models to here
# I'm adding StandardizedPullRequest here as an initial test
# for OJ-27311


@dataclass
class StandardizedPullRequest:
    pass
